from .operations import (
    create, delete, normal_filter, thread_pool, update, timeout
)

from curd import Session
from .conf import cassandra_conf

    
def create_test_table(session):
    session.execute('DROP KEYSPACE IF EXISTS curd')
    session.execute("CREATE KEYSPACE curd WITH replication = {'class': 'SimpleStrategy', 'replication_factor': 1}")
    session.execute('CREATE TABLE curd.test (id int PRIMARY KEY, text text)')
    return 'curd.test'


def test_cassandra():
    session = Session([cassandra_conf])
    create(session, create_test_table)
    update(session, create_test_table)
    delete(session, create_test_table)
    normal_filter(session, create_test_table)
    thread_pool(session, create_test_table)
    timeout(session, create_test_table)



import pytest

import copy
from multiprocessing.pool import ThreadPool
from threading import current_thread

from curd import DuplicateKeyError, OperationFailure


def create(session, create_test_table):
    collection = create_test_table(session)
    data = {'id': 100, 'text': 'test'}
    session.create(collection, data)
    with pytest.raises(DuplicateKeyError):
        session.create(collection, data, mode='insert')
    # session.create(collection, data, mode='ignore')
    d = session.get(collection, [('=', 'id', data['id'])])
    assert d == data
    data2 = copy.deepcopy(data)
    data2['text'] = 't2'
    session.create(collection, data2, mode='replace')
    d = session.get(collection, [('=', 'id', data2['id'])])
    assert d != data
    
    
def update(session, create_test_table):
    collection = create_test_table(session)
    data = {'id': 100, 'text': 'test'}
    session.create(collection, data)
    session.update(collection, {'text': 't2'}, [('=', 'id', data['id'])])
    d = session.get(collection, [('=', 'id', data['id'])])
    assert d['text'] == 't2'
    
    
def delete(session, create_test_table):
    collection = create_test_table(session)
    data = {'id': 100, 'text': 'test'}
    session.create(collection, data)
    session.delete(collection, [('=', 'id', data['id'])])
    d = session.get(collection, [('=', 'id', data['id'])])
    assert d is None
    
    
def normal_filter(session, create_test_table):
    collection = create_test_table(session)
    for i in range(1, 2000):
        session.create(collection, {'id': i, 'text': 'test'})
        
    items = session.filter(
        collection, [('<=', 'id', 1000)], fields=['text'], limit=None)
    assert len(items) == 1000
    for item in items:
        assert not item.get('id')

    
def filter_with_order_by(session, create_test_table):
    collection = create_test_table(session)
    for i in range(1, 2000):
        session.create(collection, {'id': i, 'text': 'test'})

    items = session.filter(
        collection, [('IN', 'id', (1, 32, 16))],
        fields=['id'], order_by='id')
    assert [item['id'] for item in items] == [1, 16, 32]

    items = session.filter(
        collection, [('IN', 'id', (1, 32, 16))],
        fields=['id'], order_by='-id')
    assert [item['id'] for item in items] == [32, 16, 1]
    
    
def timeout(session, create_test_table):
    collection = create_test_table(session)
    for i in range(1, 2000):
        session.create(collection, {'id': i, 'text': 'test'})

    with pytest.raises(OperationFailure):
        items = session.filter(
            collection, [('IN', 'id', (1, 32, 16))],
            fields=['id'], order_by='id', timeout=0.000001)
        print(items)


def thread_pool(session, create_test_table):
    collection = create_test_table(session)
    
    def create(i):
        session.create(
            collection,
            {'id': i, 'text': current_thread().getName()}
        )
    pool_size = 20
    pool = ThreadPool(pool_size)
    pool.map(create, range(1, 10000))
    pool.terminate()
    
    items = session.filter(collection, [('>=', 'id', 1)], limit=None)
    assert len(items) == 9999
    t_names = set()
    for item in items:
        t_names.add(item['text'])
    assert len(t_names) == pool_size



import logging

from ..errors import ProgrammingError


logger = logging.Logger('curd')

DEFAULT_FILTER_LIMIT = 100
DEFAULT_TIMEOUT = 3600 * 10

CREATE_MODE = ('INSERT', 'IGNORE', 'REPLACE')
FILTER_OP = ('<', '>', '>=', '<=', '=', '!=', 'IN')
CURD_FUNCTIONS = (
    'create', 'update', 'get', 'delete', 'filter', 'exist', 'execute'
)

OP_RETRY_WARNING = 'RETRY: {}'


class BaseConnection(object):
    def _check_filters(self, filters):
        if filters is None:
            filters = []
        new_filters = []
        for op, k, v in filters:
            if op.upper() not in FILTER_OP:
                raise ProgrammingError('Not support filter Operator')
            else:
                new_filters.append((op.upper(), k, v))
        return new_filters
    
    def create(self, collection, data, mode='INSERT', **kwargs):
        raise NotImplementedError

    def update(self, collection, data, filters, **kwargs):
        raise NotImplementedError

    def delete(self, collection, filters, **kwargs):
        raise NotImplementedError
    
    def filter(self, collection, filters=None, fields=None,
               order_by=None, limit=DEFAULT_FILTER_LIMIT, **kwargs):
        raise NotImplementedError
    
    def get(self, collection, filters=None, fields=None, **kwargs):
        rows = self.filter(collection, filters, fields, limit=1, **kwargs)
        if rows:
            return rows[0]
        else:
            return None

    def exist(self, collection, filters=None, **kwargs):
        if filters:
            fields = list(filters.keys())[0]
            data = self.get(collection, filters, fields=fields, **kwargs)
            if data:
                return False
            else:
                return True
        else:
            raise ProgrammingError('exist without filter is not supported')



class Error(Exception):
    pass


class WrappedError(Error):
    BASE_MESSAGE = 'WrappedError'

    def __init__(self, *args, origin_error=None, **kwargs):
        self._origin_error = origin_error
        if self._origin_error:
            message = self.BASE_MESSAGE + ': ' + str(self._origin_error)
        else:
            message = self.BASE_MESSAGE
        super().__init__(message, *args, **kwargs)


class ConnectError(WrappedError):
    '''
    errors when connect to database
    '''
    
    BASE_MESSAGE = 'ConnectError'


class UnexpectedError(WrappedError):
    '''
    uncategorized errors
    '''
    
    BASE_MESSAGE = 'UnexpectedError'


class OperationFailure(WrappedError):
    '''
    errors like timeout, mysql gone away, retriable
    '''
    
    BASE_MESSAGE = 'OperationFailure'


class ProgrammingError(WrappedError):
    BASE_MESSAGE = 'ProgrammingError'


class DuplicateKeyError(Error):
    pass



import copy
import pymysql

from ..errors import (
    UnexpectedError, OperationFailure, ProgrammingError,
    ConnectError,
    DuplicateKeyError
)
from . import logger
from .utils.sql import (
    query_parameters_from_create,
    query_parameters_from_update,
    query_parameters_from_delete,
    query_parameters_from_filter
)
from . import (
    BaseConnection, DEFAULT_FILTER_LIMIT, DEFAULT_TIMEOUT, OP_RETRY_WARNING
)

# https://www.briandunning.com/error-codes/?source=MySQL

PE_MYSQL_ERROR_CODE_LIST = [
    1265,  # Data truncated, tidb error
] + list(range(1046, 1076))


PE_DUPLICATE_ENTRY_KEY_ERROR_CODE = 1062

OF_MYSQL_ERROR_CODE_LIST = [0, 1040, 2006, 2013]
OF_MYSQL_RETRY_ERROR_CODE_LIST = [
    0,  # mysql interface error
    1040,  # mysql too many connections
    2006,  # mysql gone away
    2013,  # mysql connection timeout
]

TIDB_TRY_AGAIN_LATER_ERROR_CODE = 1105
OF_TIDB_ERROR_CODE_LIST = OF_MYSQL_ERROR_CODE_LIST + \
    [TIDB_TRY_AGAIN_LATER_ERROR_CODE]
OF_TIDB_RETRY_ERROR_CODE_LIST = OF_MYSQL_RETRY_ERROR_CODE_LIST + \
    [TIDB_TRY_AGAIN_LATER_ERROR_CODE]


class MysqlConnection(BaseConnection):
    pe_mysql_error_code_list = PE_MYSQL_ERROR_CODE_LIST
    pe_duplicate_entry_key_error_code = PE_DUPLICATE_ENTRY_KEY_ERROR_CODE
    
    of_mysql_error_code_list = OF_MYSQL_ERROR_CODE_LIST
    of_mysql_retry_error_code_list = OF_MYSQL_RETRY_ERROR_CODE_LIST
    
    def __init__(self, conf):
        self._conf = conf
        self.conn, self.cursor = None, None

        self.max_op_fail_retry = conf.get('max_op_fail_retry', 0)
        self.default_timeout = conf.get('timeout', DEFAULT_TIMEOUT)
    
    def _connect(self, conf):
        conf = copy.deepcopy(conf)

        self.max_op_fail_retry = conf.pop('max_op_fail_retry', 0)
        self.default_timeout = conf.pop('timeout', DEFAULT_TIMEOUT)
        
        conf['use_unicode'] = True
        conf['charset'] = 'utf8mb4'
        conf['autocommit'] = True
        
        conf['read_timeout'] = self.default_timeout
        conf['write_timeout'] = self.default_timeout
        
        if conf.pop('tidb_patch', False):
            self.patch_execute_as_tidb()

        conn = pymysql.connect(**conf)
        cursor = conn.cursor(pymysql.cursors.DictCursor)
        return conn, cursor
    
    def connect(self, conf):
        try:
            return self._connect(conf)
        except Exception as e:
            raise ConnectError(origin_error=e)
    
    def close(self):
        if self.cursor:
            try:
                self.cursor.close()
            except Exception as e:
                logger.warning(str(e))
        if self.conn:
        
            try:
                self.conn.close()
            except Exception as e:
                logger.warning(str(e))

        self.conn, self.cursor = None, None
        
    def _execute(self, query, params, timeout):
        if self.cursor and self.conn:
            pass
        else:
            self.conn, self.cursor = self.connect(self._conf)
            
        self.conn._read_timeout = timeout
        self.conn._write_timeout = timeout
        
        try:
            self.cursor.execute(query, params)
        except pymysql.err.ProgrammingError as e:
            raise ProgrammingError(origin_error=e)
        except Exception as e:
            if isinstance(e.args, tuple) and len(e.args) >= 1:
                if e.args[0] in self.pe_mysql_error_code_list:
                    raise ProgrammingError(origin_error=e)
                elif e.args[0] in self.of_mysql_error_code_list:
                    raise OperationFailure(origin_error=e)
                else:
                    raise UnexpectedError(origin_error=e)
            else:
                raise UnexpectedError(origin_error=e)
        else:
            return list(self.cursor.fetchall())
            
    def execute(self, query, params=None, retry=None, timeout=None):
        if retry is None:
            retry = self.max_op_fail_retry
            
        if timeout is None:
            timeout = self.default_timeout
            
        retry_no = 0
        while True:
            try:
                rows = self._execute(query, params, timeout)
            except OperationFailure as e:
                self.close()
                if retry_no < retry:
                    logger.warning(OP_RETRY_WARNING.format(str(e)))
                    retry_no += 1
                else:
                    raise
            except ProgrammingError:
                raise
            except (UnexpectedError, Exception, KeyboardInterrupt):
                self.close()
                raise
            else:
                return rows

    def create(self, collection, data, mode='INSERT', **kwargs):
        query, params = query_parameters_from_create(
            collection, data, mode.upper())
        try:
            self.execute(query, params, **kwargs)
        except ProgrammingError as e:
            if e._origin_error.args[0] == self.pe_duplicate_entry_key_error_code:
                raise DuplicateKeyError(str(e._origin_error))
            else:
                raise

    def update(self, collection, data, filters, **kwargs):
        filters = self._check_filters(filters)
        query, params = query_parameters_from_update(collection, filters, data)
        self.execute(query, params, **kwargs)

    def delete(self, collection, filters, **kwargs):
        filters = self._check_filters(filters)
        query, params = query_parameters_from_delete(collection, filters)
        self.execute(query, params, **kwargs)
        
    def filter(self, collection, filters=None, fields=None,
               order_by=None, limit=DEFAULT_FILTER_LIMIT, **kwargs):
        filters = self._check_filters(filters)
        query, params = query_parameters_from_filter(
            collection, filters, fields, order_by, limit)
        rows = self.execute(query, params, **kwargs)
        return rows
    
    def patch_execute_as_tidb(self):
        self.of_mysql_error_code_list = OF_TIDB_ERROR_CODE_LIST
        self.of_mysql_retry_error_code_list = OF_TIDB_RETRY_ERROR_CODE_LIST



import pymongo

from ..errors import (
    UnexpectedError, OperationFailure, ProgrammingError,
    ConnectError,
    DuplicateKeyError
)
from . import logger


class MongoConnection(object):
    def __init__(self, conf):
        self._conf = conf
        self.client = None

        self._cache_dbs = {}
        self._cache_collections = {}

    def _connect(self, conf):
        client = pymongo.MongoClient(**conf)
        return client

    def connect(self, conf):
        try:
            return self._connect(conf)
        except Exception as e:
            raise ConnectError(origin_error=e)
        
    def close(self):
        try:
            self.client.close()
        except Exception as e:
            logger.warning(str(e))
            
        self.client = None
        
    def _get_db(self, db_name):
        db = self._cache_dbs.get(db_name, None)
        if db:
            return db
        else:
            db = getattr(self.client, db_name)
            self._cache_dbs[db_name] = db
            return db
    
    def _get_collection(self, db_name, collection_name):
        key = db_name + '.' + collection_name
        collection = self._cache_collections.get(key, None)
        if collection:
            return collection
        else:
            collection = getattr(self._get_db(db_name), collection_name)
            self._cache_collections[collection_name] = collection
            return collection
        
    def _execute(self, db, collection, operation, *args, **kwargs):
        if not self.client:
            self.client = self.connect(self._conf)
    
        collection = self._get_collection(db, collection)
        func = getattr(collection, operation)
    
        try:
            return func(*args, **kwargs)
        except pymongo.errors.OperationFailure as e:
            raise OperationFailure(origin_error=e)
        except (
                pymongo.errors.DuplicateKeyError,
                pymongo.errors.InvalidOperation,
                pymongo.errors.InvalidName,
                pymongo.errors.CollectionInvalid,
                pymongo.errors.DocumentTooLarge,
        ) as e:
            raise ProgrammingError(origin_error=e)
        except Exception as e:
            raise UnexpectedError(origin_error=e)
        
    def execute(self, db, collection, operation, *args, **kwargs):
        try:
            return list(self._execute(db, collection, operation, *args, **kwargs))
        except OperationFailure:
            self.close()
            raise
        except ProgrammingError:
            raise
        except (UnexpectedError, Exception, KeyboardInterrupt):
            self.close()
            raise
        
    def create(self, collection, data, mode='INSERT', **kwargs):
        db, collection = collection.split('.', 1)
        mode = mode.upper()
        if mode in ('INSERT', 'IGNORE'):
            try:
                self.execute(db, collection, 'insert_one', data, **kwargs)
            except ProgrammingError as e:
                if isinstance(e._origin_error, pymongo.errors.DuplicateKeyError):
                    if mode == 'INSERT':
                        raise DuplicateKeyError(str(e._origin_error))
                    else:
                        pass
                else:
                    raise
        elif mode == 'REPLACE':
            m_id = data.get('_id', None)
            if m_id:
                self.execute(
                    db, collection, 'replace_one',
                    {'_id': m_id}, data, upsert=True, **kwargs
                )
            else:
                raise ProgrammingError(
                    'mongodb create mode replace, '
                    'data without _id is not supported')

    def update(self, collection, filters, data, **kwargs):
        db, collection = collection.split('.', 1)
        self.execute(
            db, collection, 'update_many',
            filters, data, **kwargs
        )

    def delete(self, collection, filters, **kwargs):
        db, collection = collection.split('.', 1)
        self.execute(
            db, collection, 'delete_many',
            filters, **kwargs
        )
        
    def filter(self, collection, filters, fields=None,
               order_by=None, limit=1000, **kwargs):
        db, collection = collection.split('.', 1)
        if fields:
            fields_dict = dict([(f, 1) for f in fields or []])
        
            if '_id' not in fields:
                fields_dict['_id'] = 0
        
            return self.execute(
                db, collection, 'find',
                filters, fields_dict, **kwargs
            )
        else:
            return self.execute(
                db, collection, 'find',
                filters, **kwargs
            )



import copy
from cassandra.auth import PlainTextAuthProvider
from cassandra.cluster import Cluster
from cassandra import Timeout, OperationTimedOut, InvalidRequest

from ..errors import (
    UnexpectedError, OperationFailure, ProgrammingError,
    ConnectError,
    DuplicateKeyError
)
from . import logger
from .utils.cql import (
    query_parameters_from_create,
    query_parameters_from_update,
    query_parameters_from_delete,
    query_parameters_from_filter,
)
from . import (
    BaseConnection, DEFAULT_FILTER_LIMIT, DEFAULT_TIMEOUT, OP_RETRY_WARNING
)


class CassandraConnection(BaseConnection):
    
    def __init__(self, conf):
        self._conf = conf
        self.cluster, self.session = None, None

        self.max_op_fail_retry = conf.get('max_op_fail_retry', 0)
        self.default_timeout = conf.get('timeout', DEFAULT_TIMEOUT)
        
    def _connect(self, conf):
        conf = copy.deepcopy(conf)

        self.max_op_fail_retry = conf.pop('max_op_fail_retry', 0)
        self.default_timeout = conf.pop('timeout', DEFAULT_TIMEOUT)

        if conf.get('username', None):
            auth_provider = PlainTextAuthProvider(
                username=conf['username'], password=conf['password']
            )
            cluster = Cluster(conf['hosts'], auth_provider=auth_provider)
        else:
            cluster = Cluster(conf['hosts'])
        session = cluster.connect()
        return cluster, session
    
    def connect(self, conf):
        try:
            return self._connect(conf)
        except Exception as e:
            raise ConnectError(origin_error=e)

    def close(self):
        if self.session:
            try:
                self.session.shutdown()
            except Exception as e:
                logger.warning(str(e))
        if self.cluster:
            try:
                self.cluster.shutdown()
            except Exception as e:
                logger.warning(str(e))

        self.cluster, self.session = None, None
        
    def _execute(self, query, params, **kwargs):
        if self.cluster and self.session:
            pass
        else:
            self.cluster, self.session = self.connect(self._conf)
        
        try:
            result = list(self.session.execute(query, params, **kwargs))
        except (Timeout, OperationTimedOut) as e:
            raise OperationFailure(origin_error=e)
        except InvalidRequest as e:
            raise ProgrammingError(origin_error=e)
        except Exception as e:
            raise UnexpectedError(origin_error=e)
        else:
            return result
        
    def execute(self, query, params=None, retry=None, timeout=None):
        if retry is None:
            retry = self.max_op_fail_retry
            
        if timeout is None:
            timeout = self.default_timeout

        retry_no = 0
        while True:
            try:
                rows = self._execute(query, params, timeout=timeout)
            except OperationFailure as e:
                self.close()
                if retry_no < retry:
                    logger.warning(OP_RETRY_WARNING.format(str(e)))
                    retry_no += 1
                else:
                    raise
            except ProgrammingError:
                    raise
            except (UnexpectedError, Exception, KeyboardInterrupt):
                self.close()
                raise
            else:
                return [row._asdict() for row in rows]
        
    def create(self, collection, data, mode='INSERT', **kwargs):
        query, params = query_parameters_from_create(
            collection, data, mode.upper())
        rows = self.execute(query, params, **kwargs)
        if rows and mode.upper() != 'IGNORE' and not rows[0].get('applied', True):
            raise DuplicateKeyError

    def update(self, collection, data, filters, **kwargs):
        filters = self._check_filters(filters)
        query, params = query_parameters_from_update(collection, filters, data)
        self.execute(query, params, **kwargs)
        
    def delete(self, collection, filters, **kwargs):
        filters = self._check_filters(filters)
        query, params = query_parameters_from_delete(collection, filters)
        self.execute(query, params, **kwargs)
        
    def filter(self, collection, filters=None, fields=None,
               order_by=None, limit=DEFAULT_FILTER_LIMIT, **kwargs):
        filters = self._check_filters(filters)
        query, params = query_parameters_from_filter(
            collection, filters, fields, order_by, limit)
        rows = self.execute(query, params, **kwargs)
        return rows



from datetime import datetime, timezone


class BaseClause(object):
    def __init__(self, field, value):
        self._field = field
        self._value = value
        
    @property
    def field(self):
        return '.'.join(
            ['`{}`'.format(i) for i in self._field.replace('`', '').split('.')])
    
    @property
    def value(self):
        if isinstance(self._value, datetime) and self._value.tzinfo:
            return self._value.astimezone(tz=timezone.utc).replace(tzinfo=None)
        elif isinstance(self._value, list) or isinstance(self._value, tuple):
            value = []
            for v in self._value:
                if isinstance(self._value, datetime) and self._value.tzinfo:
                    value.append(
                        v.astimezone(tz=timezone.utc).replace(tzinfo=None)
                    )
                else:
                    value.append(v)
            return value
        else:
            return self._value
        
        
class WhereClause(BaseClause):
    def __init__(self, field, operator, value):
        super().__init__(field, value)
        self._operator = operator
        
    @property
    def operator(self):
        return self._operator
        
        
class FieldClause(BaseClause):
    def __init__(self, field):
        self._field = field

    
class AssignmentClause(BaseClause):
    pass


class BaseSQLStatement(object):
    def __init__(self):
        self.query = None
        self.params = []
    
    def generate_query_field(self, table):
        return table.field
    
    def generate_query_where(self, where):
        query = ''
        
        if where:
            query += 'WHERE '
            segs = []
            for where_clause in where:
                if where_clause.operator == 'IN':
                    value_count = len(where_clause.value)
                    segs.append(
                        '{} {} {}'.format(
                            where_clause.field,
                            where_clause.operator,
                            '({})'.format(', '.join(['%s']*value_count))
                        )
                    )
                    
                    for v in where_clause.value:
                        self.params.append(v)
                else:
                    segs.append(
                        '{} {} {}'.format(
                            where_clause.field, where_clause.operator, '%s'
                        )
                    )
                    self.params.append(where_clause.value)
            query += ' AND '.join(segs)
        return query

    def generate_query_fields(self, fields):
        if fields:
            return ', '.join([field_clause.field for field_clause in fields])
        else:
            return '*'
        
    def generate_query_limit(self, limit):
        if limit:
            return 'LIMIT {}'.format(limit)
        else:
            return ''
        
    def generate_query_order_by(self, order_by):
        query = ''
        if order_by:
            query += 'ORDER BY '
            segs = []
            for field_clause in order_by:
                if field_clause._field.startswith('-'):
                    field_clause._field = field_clause._field[1:]
                    seg = field_clause.field + ' DESC'
                else:
                    seg = field_clause.field
                segs.append(seg)
            query += ', '.join(segs)
        return query


class SelectStatement(BaseSQLStatement):
    BASE_QUERY = 'SELECT {} FROM {} {}'
    
    def __init__(self, table, fields=None, where=None, order_by=None, limit=None):
        super().__init__()

        self.table = table
        self.fields = fields
        self.where = where
        self.order_by = order_by
        self.limit = limit
        
    def as_sql(self):
        query_table = self.generate_query_field(self.table)
        query_fields = self.generate_query_fields(self.fields)
        query_where = self.generate_query_where(self.where)
        query_order_by = self.generate_query_order_by(self.order_by)
        query_limit = self.generate_query_limit(self.limit)
    
        extra_query = ' '.join([
            i for i in [query_where, query_order_by, query_limit] if i
        ])

        self.query = self.BASE_QUERY.format(
            query_fields, query_table, extra_query
        )

        return self.query, self.params
    
    
class DeleteStatement(BaseSQLStatement):
    BASE_QUERY = 'DELETE FROM {} {}'

    def __init__(self, table, where=None):
        super().__init__()
        self.table = table
        self.where = where
    
    def as_sql(self):
        query_table = self.generate_query_field(self.table)
        query_where = self.generate_query_where(self.where)
        extra_query = query_where
    
        self.query = self.BASE_QUERY.format(
            query_table, extra_query
        )
        return self.query, self.params
    
    
class CreateStatement(BaseSQLStatement):
    BASE_QUERY = '{} INTO {} ({}) VALUES ({})'
    
    def __init__(self, table, assignments, mode):
        super().__init__()
        self.table = table
        self.assignments = assignments
        self.mode = mode
        
    def generate_query_mode(self, mode):
        if mode == 'INSERT':
            return 'INSERT'
        elif mode == 'IGNORE':
            return 'INSERT IGNORE'
        elif mode == 'REPLACE':
            return 'REPLACE'
        
    def generate_query_fields_values(self, assignments):
        query_fields = ', '.join([a.field for a in assignments])
        query_values = ', '.join(['%s']*len(assignments))
        for a in assignments:
            self.params.append(a.value)
        return query_fields, query_values
    
    def as_sql(self):
        
        query_mode = self.generate_query_mode(self.mode)
        
        query_table = self.generate_query_field(self.table)
        
        query_fields, query_values = self.generate_query_fields_values(
            self.assignments)
        
        self.query = self.BASE_QUERY.format(
            query_mode, query_table, query_fields, query_values
        )
        return self.query, self.params
    
    
class UpdateStatement(BaseSQLStatement):
    BASE_QUERY = 'UPDATE {} SET {} {}'
    
    def __init__(self, table, assignments, where=None):
        super().__init__()
        self.table = table
        self.assignments = assignments
        self.where = where
        
    def generate_query_fields_values(self, assignments):
        query = ', '.join(
            [a.field + '=%s' for a in assignments]
        )
        for a in assignments:
            self.params.append(a.value)
        return query

    def as_sql(self):
        query_table = self.generate_query_field(self.table)
    
        query_fields_values = self.generate_query_fields_values(
            self.assignments)

        query_where = self.generate_query_where(self.where)
    
        self.query = self.BASE_QUERY.format(
            query_table, query_fields_values, query_where
        )
        return self.query, self.params
    

def where_clauses_from_filters(filters):
    where_clauses = []
    for op, k, v in filters:
        where_clause = WhereClause(k, op, v)
        where_clauses.append(where_clause)
    return where_clauses


def assignment_clauses_clauses_from_filters(data):
    assignment_clauses = []
    for k, v in data.items():
        clause = AssignmentClause(k, v)
        assignment_clauses.append(clause)
    return assignment_clauses


def query_parameters_from_create(collection, data, mode='INSERT'):
    table = FieldClause(collection)
    assignments = assignment_clauses_clauses_from_filters(data)
    query, params = CreateStatement(table, assignments, mode).as_sql()
    return query, params


def query_parameters_from_update(collection, filters, data):
    table = FieldClause(collection)
    assignments = assignment_clauses_clauses_from_filters(data)
    where = where_clauses_from_filters(filters)
    query, params = UpdateStatement(table, assignments, where).as_sql()
    return query, params


def query_parameters_from_get(collection, filters, fields=None):
    table = FieldClause(collection)
    where = where_clauses_from_filters(filters)
    fields = [FieldClause(f) for f in (fields or [])]
    query, params = SelectStatement(table, fields, where, limit=1).as_sql()
    return query, params


def query_parameters_from_delete(collection, filters):
    table = FieldClause(collection)
    where = where_clauses_from_filters(filters)
    query, params = DeleteStatement(table, where).as_sql()
    return query, params


def query_parameters_from_filter(
        collection, filters, fields=None, order_by=None, limit=None):
    table = FieldClause(collection)
    where = where_clauses_from_filters(filters)
    fields = [FieldClause(f) for f in (fields or [])]
    
    if order_by is None:
        order_by = []
    elif isinstance(order_by, str):
        order_by = [order_by]
    order_by = [FieldClause(f) for f in order_by]
    
    query, params = SelectStatement(
        table, fields, where, order_by, limit).as_sql()
    return query, params



import json
import queue
from functools import partial
from collections import OrderedDict

from .connections import CURD_FUNCTIONS
from .connections.mysql import MysqlConnection
from .connections.cassandra import CassandraConnection

from .errors import ProgrammingError


class CassandraConnectionPool(CassandraConnection):
    pass


class MysqlConnectionPool(object):
    def __init__(self, *args, **kwargs):
        self.conn_queue = queue.Queue()
        self._args = args
        self._kwargs = kwargs
        
        for func in CURD_FUNCTIONS:
            setattr(self, func, partial(self._wrap_func, func))
            
    def get_connection(self):
        return MysqlConnection(*self._args, **self._kwargs)
        
    def _wrap_func(self, func, *args, **kwargs):
        try:
            conn = self.conn_queue.get_nowait()
        except queue.Empty:
            conn = self.get_connection()
            
        try:
            return getattr(conn, func)(*args, **kwargs)
        except:
            raise
        finally:
            self.conn_queue.put_nowait(conn)
            
    def close(self):
        while True:
            try:
                conn = self.conn_queue.get_nowait()
            except queue.Empty:
                break
            else:
                conn.close()


class Session(object):
    
    '''
    mysql db conf
    {
        'type': 'mysql'
        'conf': {
            'host': '127.0.0.1',
            'port': 3306,
            'user': 'user',
            'password': 'password',
            'max_op_fail_retry': 3,
            'timeout': 60
        }
    }
    tidb conf
    {
        'type': 'mysql'
        'conf': {
            'host': '127.0.0.1',
            'port': 3306,
            'user': 'user',
            'password': 'password',
            'tidb_patch': True,
            'max_op_fail_retry': 3,
            'timeout': 60
        }
    }
    cassandra db conf
    {
        'type': 'cassandra',
        'conf': {
            'hosts': ['127.0.0.1'],
            'username': 'username',
            'password': 'password',
            'max_op_fail_retry': 3,
            'timeout': 60
        }
    }
    '''
    
    def __init__(self, dbs=None):
        self._connection_cache = OrderedDict()
        self._default_connection = None
        
        if dbs:
            for db in dbs:
                self._get_connection(db)
        
    def _create_connection(self, db):
        if db['type'] == 'mysql':
            return MysqlConnectionPool(db['conf'])
        elif db['type'] == 'cassandra':
            return CassandraConnectionPool(db['conf'])
        else:
            raise ProgrammingError('not supported database')
        
    def set_default_connection(self, db):
        key = json.dumps(db)
        conn = self._connection_cache.get(key, None)
        if not conn:
            conn = self._create_connection(db)
            self._connection_cache[key] = conn
        self._default_connection = conn
        
    def _get_connection(self, db):
        key = json.dumps(db)
        conn = self._connection_cache.get(key, None)
        if conn:
            return conn
        else:
            conn = self._create_connection(db)
            self._connection_cache[key] = conn
            
            if not self._default_connection:
                self._default_connection = conn
            
            return conn
        
    def using(self, db=None):
        if db:
            return self._get_connection(db)
        else:
            return self._default_connection
        
    def __getattr__(self, item):
        if item in CURD_FUNCTIONS:
            if self._default_connection:
                return getattr(self._default_connection, item)
            else:
                raise ProgrammingError('no database conf')
        else:
            raise AttributeError
            
    def close(self):
        for k, v in self._connection_cache.items():
            v.close()
        self._connection_cache = OrderedDict()
        self._default_connection = None



from .operations import (
    create, delete, normal_filter, filter_with_order_by, thread_pool, update
)

from curd import Session
from .conf import mysql_conf

    
def create_test_table(session):
    session.execute('DROP DATABASE IF EXISTS `curd`')
    session.execute('CREATE DATABASE `curd` DEFAULT CHARACTER SET utf8mb4')
    session.execute('CREATE TABLE `curd`.`test` (`id` int(11) unsigned NOT NULL AUTO_INCREMENT, `text` text, PRIMARY KEY (`id`)) ENGINE=InnoDB AUTO_INCREMENT=300000 DEFAULT CHARSET=utf8mb4')
    return 'curd.test'


def test_mysql():
    session = Session([mysql_conf])
    create(session, create_test_table)
    update(session, create_test_table)
    delete(session, create_test_table)
    normal_filter(session, create_test_table)
    filter_with_order_by(session, create_test_table)
    thread_pool(session, create_test_table)



from cassandra.cqlengine.statements import (
    InsertStatement,
    UpdateStatement,
    SelectStatement as _SelectStatement,
    DeleteStatement,
    WhereClause, AssignmentClause,
    six
)
from cassandra.cqlengine.operators import BaseWhereOperator


class SelectStatement(_SelectStatement):
    def __unicode__(self):
        qs = ['SELECT']
        if self.distinct_fields:
            if self.count:
                qs += ['DISTINCT COUNT({0})'.format(', '.join(['"{0}"'.format(f) for f in self.distinct_fields]))]
            else:
                qs += ['DISTINCT {0}'.format(', '.join(['"{0}"'.format(f) for f in self.distinct_fields]))]
        elif self.count:
            qs += ['COUNT(*)']
        else:
            qs += [', '.join(['"{0}"'.format(f) for f in self.fields]) if self.fields else '*']
        qs += ['FROM', self.table]

        if self.where_clauses:
            qs += [self._where]

        if self.order_by and not self.count:
            segs = []
            for o in self.order_by:
                if o.startswith('-'):
                    segs.append(six.text_type(o[1:]) + ' DESC')
                else:
                    segs.append(six.text_type(o))
            
            qs += ['ORDER BY {0}'.format(', '.join(segs))]

        if self.limit:
            qs += ['LIMIT {0}'.format(self.limit)]

        if self.allow_filtering:
            qs += ['ALLOW FILTERING']

        return ' '.join(qs)


CQL_SYMBOL = dict(
    [(op.cql_symbol, op.symbol) for op in BaseWhereOperator.opmap.values()]
)


def where_clauses_from_filters(filters):
    where_clauses = []
    for op, k, v in filters:
        where_clause = WhereClause(
            k, BaseWhereOperator.get_operator(CQL_SYMBOL[op.upper()])(), v)
        where_clauses.append(where_clause)
    return where_clauses


def assignment_clauses_from_data(data):
    assignment_clauses = []
    for k, v in data.items():
        assignment_clauses.append(AssignmentClause(k, v))
    return assignment_clauses


def query_parameters_from_create(collection, data, mode='INSERT'):
    assignment_clauses = assignment_clauses_from_data(data)
    if mode == 'REPLACE':
        statement = InsertStatement(
            table=collection, assignments=assignment_clauses,
        )
    else:
        statement = InsertStatement(
            table=collection, assignments=assignment_clauses,
            if_not_exists=True
        )
    return str(statement), statement.get_context()


def query_parameters_from_update(collection, filters, data):
    where_clauses = where_clauses_from_filters(filters)
    assignment_clauses = assignment_clauses_from_data(data)
    statement = UpdateStatement(
        table=collection, assignments=assignment_clauses,
        where=where_clauses, if_exists=True
    )
    return str(statement), statement.get_context()


def query_parameters_from_delete(collection, filters):
    where_clauses = where_clauses_from_filters(filters)
    statement = DeleteStatement(
        table=collection, where=where_clauses
    )
    return str(statement), statement.get_context()


def query_parameters_from_filter(
    collection, filters, fields=None, order_by=None, limit=None
):
    where_clauses = where_clauses_from_filters(filters)
    statement = SelectStatement(
        table=collection, fields=fields,
        where=where_clauses, allow_filtering=True,
        order_by=order_by, limit=limit
    )
    return str(statement), statement.get_context()



from setuptools import setup, find_packages


setup(
    name='curd',
    version='0.0.8.1',
    url='https://github.com/jdxin0/curd',
    description='db operations',
    author='jdxin0',
    author_email='jdxin00@gmail.com',
    license='MIT',
    keywords='db operations',
    packages=find_packages(),
    install_requires=[
        'cassandra-driver==3.11.0',
        'PyMySQL==0.7.11',
    ],
)